import typing
from urllib import parse

from wikitools import console, reference_parser


class Brackets():
    # Helper class keeping track of when brackets open and close
    def __init__(self, left: str, right: str):
        self.left = left
        self.right = right
        self.depth = 0

    left: str
    right: str
    depth: int

    def closed(self, c: str):
        if c == self.left:
            self.depth += 1
        elif c == self.right:
            self.depth -= 1
        if self.depth == 0:
            return True
        return False


class State:
    IDLE = 0
    START = 1
    INLINE = 2
    REFERENCE = 3


class Link(typing.NamedTuple):
    """
    A Markdown link, inline- or reference-style, external or internal.
    May be relative. Example:

        See [Difficulty Names](/wiki/Beatmap/Difficulty#naming-conventions)

    - alt_text: 'Difficulty Names'
    - raw_location: '/wiki/Beatmap/Difficulty#naming-conventions'
    - parsed_location: urllib.parse.ParseResult with all of its fields

    Another example:

        ![Player is AFK](img/chat-console-afk.png "Player is away from keyboard")

    - alt_text: 'Player is AFK'
    - raw_location: 'img/chat-console-afk.png'
    - parsed_location: urllib.parse.ParseResult with all of its fields
    - title: ' "Player is away from keyboard"'
    """

    # Link position within the line. Example:
    #   See also: [Difficulty names](/wiki/Beatmap/Difficulty#naming-conventions)
    #             ^ start                                                       ^ end
    start: int
    end: int

    # Sections of a link. Example:
    #    ![Player is AFK](img/chat-console-afk.png "Player is away from keyboard")
    #      ^ alt_text  ^
    #                     ^ ----- location ----- ^
    #                                             ^ ---------- title ---------- ^
    #                     ^ --------------------- content --------------------- ^
    #     ^ ------------------ full_link / full_coloured_link ------------------ ^
    alt_text: str
    raw_location: str
    parsed_location: parse.ParseResult
    title: str

    @property
    def content(self):
        return self.raw_location if not self.title else f"{self.raw_location} {self.title}"

    @property
    def full_link(self):
        if self.is_reference:
            return f"[{self.alt_text}][{self.content}]"
        else:
            return f"[{self.alt_text}]({self.content})"

    @property
    def fragment_start(self):
        """
        Position of the link #fragment in the line, if there is one. Otherwise, the same value as the end of the link.
        """
        return self.start + len(self.alt_text) + 2 + len(self.parsed_location.path) + 1

    def colourise_link(self, fragment_only=False):
        return "{alt_text_in_braces}{left_brace}{location}{extra}{right_brace}".format(
            alt_text_in_braces=console.green(f"[{self.alt_text}]"),
            left_brace=console.green('[') if self.is_reference else console.green('('),
            location=self.colourise_location(fragment_only=fragment_only),
            extra=" " + console.blue(self.title) if self.title else "",
            right_brace=console.green(']') if self.is_reference else console.green(')'),
        )

    def colourise_location(self, fragment_only=False):
        if fragment_only:
            return "".join((
                console.green(self.parsed_location.path),
                console.red('#' + self.parsed_location.fragment)
            ))
        return console.red(self.raw_location)

    def resolve(
        self, references: reference_parser.References
    ) -> typing.Optional[reference_parser.Reference]:
        if not self.is_reference:
            return None
        return references.get(self.parsed_location.path)

    # Whether the link is a reference-style link. The only difference is that
    # `location` is a reference and needs to be resolved later.
    #
    # The syntax for such links is the almost the same as regular links:
    #    [text][reference]
    #
    # The reference can then later be defined at the start of a new line:
    #    [reference]: link
    is_reference: bool


def find_link(s: str, index=0) -> typing.Optional[Link]:
    """
    Finds the first valid Markdown link found in the string `s`, starting the search from position `index`.
    The following are considered links (title and alt text may be omitted):
        - [alt_text](/loca/ti/on "Title")
        - ![alt_text](/path/to/image "Title), with ! not being considered a part of the link
        - [artist - title (creator) [diff]](/loca/ti/ion_(with_parentheses) "Title")
        - [alt_text][reference], with exact locations found separately via find_reference()
    """

    state = State.IDLE

    start = -1
    location = -1
    extra = None
    end = None

    parens = Brackets('(', ')')
    brackets = Brackets('[', ']')

    for i in range(index, len(s)):
        c = s[i]
        if state == State.IDLE and c == '[':
            # potential start of a link
            brackets.depth += 1
            state = State.START
            start = i
            continue

        if state == State.START:
            if brackets.closed(c):
                if len(s) <= i + 1:
                    # end of the line
                    state = State.IDLE
                    continue

                if s[start + 1] == '^':
                    # found a footnote -> ignore
                    state = State.IDLE
                    continue

                # the end of a bracket. the link may continue
                # to be inline- or reference-style
                if s[i + 1] == '(':
                    state = State.INLINE
                    location = i + 2
                elif s[i + 1] == '[':
                    state = State.REFERENCE
                    location = i + 2
                else:
                    state = State.IDLE
            continue

        if state == State.INLINE:
            if c == ' ':
                if extra is None:
                    # start of extra part
                    extra = i

            if parens.closed(c):
                # end of a complete link
                end = i
                if extra is None:
                    extra = end

                raw_location = s[location: extra]
                return Link(
                    raw_location=raw_location,
                    parsed_location=parse.urlparse(raw_location),
                    alt_text=s[start + 1: location - 2],
                    title=s[extra: end],
                    start=start,
                    end=end,
                    is_reference=False
                )
            continue

        if state == State.REFERENCE:
            if brackets.closed(c):
                # end of a complete reference-style link
                end = i
                raw_location = s[location: end]
                return Link(
                    raw_location=raw_location,
                    parsed_location=parse.urlparse(raw_location),
                    alt_text=s[start + 1: location - 2],
                    title="",
                    start=start,
                    end=end,
                    is_reference=True
                )
            continue

    return None


def find_links(line: str) -> typing.List[Link]:
    """
    Iteratively extract all links from a line.
    """

    results = []
    index = 0
    match = find_link(line, index)
    while match:
        results.append(match)
        match = find_link(line, match.end + 1)
    return results



import typing
import re


class CodeBlock(typing.NamedTuple):
    """
    A Markdown code block (monospace font + gray backround). Could be one of the following:
        - Inline: `text` or `` text `` (the latter is used for snippets containing `)
        - Multiline:
            ```
            these can start and end with more backticks,
            at which point they can contain code blocks
            inside themselves, which are ignored
            ```

    Examples:

        ```
        test test test
        ```
        start and end are both -1 for all 3 lines

        `this is a very long phrase`
        ^ start                    ^ end

    Similarly to HTML comments, links in code blocks are skipped
    """

    start: int  # 0-based position of the first character of the block's opening tag
    end: int  # 0-based position of the last character of the block's closing tag

    @property
    def is_multiline(self):
        return self.start == -1 and self.end == -1

    def contains(self, pos: int):
        return self.is_multiline or (self.start < pos and self.end > pos)


class CodeBlockParser:
    def __init__(self):
        self.__in_multiline = False
        self.__multiline_tag = ''

    @property
    def in_multiline(self) -> bool:
        return self.__in_multiline

    def count_tag_length(self, line: str, pos: int) -> int:
        tag_len = 0
        while pos + tag_len < len(line) and line[pos + tag_len] == '`':
            tag_len += 1
        return tag_len

    def parse(self, line: str) -> typing.List[CodeBlock]:
        blocks: typing.List[CodeBlock] = []

        if line.startswith('```'):
            if self.__in_multiline:
                if line.startswith(self.__multiline_tag):
                    # multiline block closed with the correct tag
                    self.__in_multiline = False
                    return [CodeBlock(start=-1, end=-1)]
            else:
                # opening a multiline block
                self.__multiline_tag = '`' * self.count_tag_length(line, 0)
                self.__in_multiline = True
                return [CodeBlock(start=-1, end=-1)]


        if self.__in_multiline:
            return [CodeBlock(start=-1, end=-1)]

        i = 0
        while i < len(line):
            if line[i] != '`':
                i += 1
                continue

            tag_len = self.count_tag_length(line, i)

            # the next tag of the same length will close the block
            closing_tag = re.search(
                    f"(?<!`){'`' * tag_len}(?!`)",
                    line[i + tag_len:]
            )

            if closing_tag:
                closing_tag_pos = closing_tag.start() + i + tag_len
                blocks.append(CodeBlock(start=i, end=closing_tag_pos + tag_len - 1))
                i = closing_tag_pos + tag_len
            else:
                # the tag wasn't closed, but there could be more code blocks
                i += tag_len

        return blocks


def is_in_code_block(link_start: int, code_blocks: typing.List[CodeBlock]) -> bool:
    if (
        not code_blocks or
        link_start < code_blocks[0].start or
        link_start > code_blocks[-1].end
    ):
        return False

    for block in code_blocks:
        if (block.contains(link_start)):
            return True

    return False



#!/usr/bin/env python3

import argparse
import os
import sys
import typing

import yamllint.cli  # type: ignore
import yamllint.config  # type: ignore
import yamllint.linter  # type: ignore
import yamllint.rules  # type: ignore

from wikitools import console
from wikitools import yaml_rules

FRONT_MATTER_DELIMITER = "---"
MARKDOWN_EXTENSION = ".md"

DEFAULT_CONFIG_CONTENT = 'extends: default'


def install_custom_checks(config: yamllint.config.YamlLintConfig):
    for rule_cls in yaml_rules.OSU_WIKI_RULES:
        config.rules[rule_cls.ID] = yamllint.rules._RULES[rule_cls.ID] = rule_cls()


def front_matter(fileobj: typing.TextIO, filepath: str) -> str:
    delimiters = 0
    lines = []
    for line in fileobj:
        if line.split("#")[0].strip() == FRONT_MATTER_DELIMITER:
            delimiters += 1
        # Stop on second delimiter, or when it's clear there won't be front matter at all
        if delimiters == 2 or line.startswith("# "):
            break
        lines.append(line)

    if delimiters == 0:
        return ""
    if delimiters == 2:
        return "".join(lines)
    raise ValueError(f"The file {filepath} has malformed front matter")


def read_yaml(filepath: str):
    with open(filepath, "r", encoding='utf-8') as fd:
        if filepath.endswith(MARKDOWN_EXTENSION):
            return front_matter(fd, filepath)
        else:
            return fd.read()


def file_iterator(roots: list, config: yamllint.config.YamlLintConfig):
    for item in roots:
        if os.path.isdir(item):
            for root, _, filenames in os.walk(item):
                for f in filenames:
                    filepath = os.path.join(root, f)
                    if (
                        not config.is_file_ignored(filepath) and
                        (config.is_yaml_file(filepath) or filepath.endswith(MARKDOWN_EXTENSION))
                    ):
                        yield filepath
        elif os.path.isfile(item):
            yield item


def parse_args(args: list):
    parser = argparse.ArgumentParser(
        description="Check YAML files and front matter of the Markdown documents for common mistakes"
    )
    parser.add_argument("--config", help="path to the yamllint config", default="./.yamllint.yaml")
    parser.add_argument(
        "--target", nargs="+", default=(".",),
        help="files or directories to lint (by default, assumes current working directory)"
    )
    parser.add_argument("--format", choices=("parsable", "standard", "colored", "github", "auto"),
        default="colored", help="output format (see yamllint docs for details)"
    )
    return parser.parse_args(args)


def main(*args):
    args = parse_args(args)
    if os.path.exists(args.config):
        config = yamllint.config.YamlLintConfig(file=os.path.expanduser(args.config))
    else:
        config = yamllint.config.YamlLintConfig(DEFAULT_CONFIG_CONTENT)

    install_custom_checks(config)

    max_level = 0
    for path in file_iterator(args.target, config):
        try:
            path = path[2:] if path.startswith('./') else path
            payload = read_yaml(path)
            problems = yamllint.linter.run(payload, config, path)
        except EnvironmentError as e:
            print(e, file=sys.stderr)
            sys.exit(-1)
        current_level = yamllint.cli.show_problems(problems, path, args_format=args.format, no_warn=False)
        max_level = max(max_level, current_level)

    if max_level == yamllint.linter.PROBLEM_LEVELS["error"]:
        sys.exit(1)

    print(f"{console.grey('Notice:')} No errors in YAML files detected.")


if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))



import typing


class Comment(typing.NamedTuple):
    """
    An HTML comment in a line.

    These mark regions where parsed content should be discarded.

    Example:
        <!-- this is a comment -->
        ^ start                  ^ end

    Multiline comments are expressed by setting the start and/or end
    values to -1, indicating continuation of a comment from a
    previous line or to subsequent lines respectively.

    Examples:

        A multiline comment continuing from a previous line -->
        (start = -1)                                          ^ end

        <!-- A multiline comment continuing off to subsequent lines
        ^ start                                          (end = -1)

        A whole line marked as part of a multiline comment
        (start = -1)                            (end = -1)
    """

    start: int
    end: int

    @property
    def is_multiline(self):
        return self.end == -1


class CommentParser:
    def __init__(self):
        self.__in_multiline = False

    @property
    def in_multiline(self) -> bool:
        return self.__in_multiline

    def parse(self, line: str) -> typing.List[Comment]:
        comments: typing.List[Comment] = []
        index = 0
        start = -1

        while True:
            # don't start a comment if already in one
            if not self.__in_multiline:
                start = line.find("<!--", index)
                if start == -1:
                    break  # no more comments

            end = line.find("-->", 0 if start == -1 else start)
            if end != -1:
                # found the end of a comment
                if self.__in_multiline:
                    # end of a multiline comment
                    comments.append(Comment(start=-1, end=end + 2))
                    self.__in_multiline = False
                else:
                    # whole inline comment
                    comments.append(Comment(start=start, end=end + 2))
                index = end + 3
                continue

            elif start == -1:
                # no comment start or end; the whole line is part of a comment
                comments.append(Comment(start=-1, end=-1))
                break

            else:
                # unmatched comment start: continuing to subsequent lines
                comments.append(Comment(start=start, end=-1))
                break

        self.__in_multiline = comments and comments[-1].is_multiline
        return comments


def is_in_comment(index: int, comments: typing.List[Comment]) -> bool:
    for comment in comments:
        left_bound = comment.start
        right_bound = comment.end
        if comment.start == -1:
            left_bound = float('-inf')  # type: ignore
        if comment.end == -1:
            right_bound = float('inf')  # type: ignore

        if index >= left_bound and index <= right_bound:
            return True

    return False



def red(s):
    return f"\x1b[31m{s}\x1b[0m"


def green(s):
    return f"\x1b[32m{s}\x1b[0m"


def blue(s):
    return f"\x1b[34m{s}\x1b[0m"


def yellow(s):
    return f"\x1b[33m{s}\x1b[0m"


def grey(s):
    return f"\x1b[90m{s}\x1b[0m"



#!/usr/bin/env python3

"""
This script does the following:

1) Validates commit hashes of edited outdated translations (or all translations with --all)
2) Obtains directories of the modified en.md files, then lists translations in them which:
    * are not modified in the PR (assuming that the PR author took care of the rest), and
    * don't have the outdated markers/hashes

Alternatively, it can be run with --autofix (and --autocommit, or just -fc) to outdate the translations automatically.
"""

import argparse
import os
import sys

from wikitools import article_parser, console, git_utils, file_utils

# A pull request string which disables the check (has no effect here, listed for informational purposes only)
PULL_REQUEST_TAG = "SKIP_OUTDATED_CHECK"

# The front matter tag which means the translation needs to be updated
OUTDATED_TRANSLATION_TAG = "outdated_translation"

# The front matter tag which contains the commit hash since which the translation is not up to date
OUTDATED_HASH_TAG = "outdated_since"

# The script flag which will automatically outdate translations
AUTOFIX_FLAG = "--autofix"
AUTOFIX_FLAG_SHORT = "-f"

# The script flag which will automatically commit changes
AUTOCOMMIT_FLAG = "--autocommit"
AUTOCOMMIT_FLAG_SHORT = "-c"


def print_translations_to_outdate(*filenames, outdated_hash=None, workflow=False):
    print(f"{console.red('Error:')} You have edited some original articles (en.md), but did not outdate their translations:")
    print("\n".join(console.red(f"* {filename}") for filename in filenames))
    print(f"\nIf your changes DON'T NEED to be added to the translations, add {console.red(PULL_REQUEST_TAG)} anywhere in the description of your pull request.")
    if not workflow:
        print(
            f"Otherwise, rerun the script with {console.green(AUTOFIX_FLAG)} (and perhaps {console.green(AUTOCOMMIT_FLAG)}), or "
            "add the following to each article's front matter (https://osu.ppy.sh/wiki/en/Article_styling_criteria/Formatting#front-matter):"
        )
    else:
        print(
            f"Otherwise, add the following to each article's front matter (https://osu.ppy.sh/wiki/en/Article_styling_criteria/Formatting#front-matter):"
        )
    print()
    outdated_block = "" if outdated_hash is None else f"\n{OUTDATED_HASH_TAG}: {outdated_hash}"
    front_matter = f"---{outdated_block}\n{OUTDATED_TRANSLATION_TAG}: true\n---"

    # the seemingly needless complication here ensures that the github action log displays all three lines in colour
    print("\n".join([console.green(line) for line in front_matter.split("\n")]))


def print_bad_hash_error(*filenames, outdated_hash=None):
    print(
        "{} The following translations are incorrectly outdated (the {} hash is invalid).{}".format(
            console.red("Error:"),
            console.red(OUTDATED_HASH_TAG),
            "" if outdated_hash is None else " Did you mean to use {} instead?".format(
                console.green(f"{OUTDATED_HASH_TAG}: {outdated_hash}")
            )
        )
    )
    print("\n".join(console.red(f"* {filename}") for filename in filenames))


def parse_args(args):
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawTextHelpFormatter, usage="%(prog)s check-outdated-articles [options]")
    parser.add_argument("-b", "--base-commit", help="commit since which to look for changes")
    parser.add_argument("-o", "--outdated-since", default="", help=f"commit hash for the {OUTDATED_HASH_TAG} tag, uses the value of --base-commit if unspecified")
    parser.add_argument("-a", "--all", default=False, action="store_true", help="look for incorrect hashes in all outdated articles")
    parser.add_argument(f"{AUTOFIX_FLAG_SHORT}", f"{AUTOFIX_FLAG}", default=False, action="store_true", help=f"automatically add `{OUTDATED_HASH_TAG}: {{hash}}` to outdated articles")
    parser.add_argument(f"{AUTOCOMMIT_FLAG_SHORT}", f"{AUTOCOMMIT_FLAG}", default=False, action="store_true", help=f"automatically commit changes")
    parser.add_argument("-r", "--root", help="specify repository root, current working directory assumed otherwise")
    parser.add_argument("--workflow", action='store_true', help="whether the script is run from a github workflow")
    return parser.parse_args(args)


def list_outdated_translations(all_translations, modified_translations):
    """
    List translations that are not touched by the changes and don't have the outdated tags (while they should).
    """

    for article_file in all_translations:
        if article_file in modified_translations:
            continue

        with open(article_file, "r", encoding='utf-8') as fd:
            front_matter = article_parser.load_front_matter(fd)
        if OUTDATED_HASH_TAG in front_matter or front_matter.get(OUTDATED_TRANSLATION_TAG, False):
            continue

        yield article_file


def list_modified_translations(base_commit):
    return filter(file_utils.is_translation, git_utils.git_diff('wiki/**/*.md', base_commit=base_commit))


def list_modified_originals(base_commit):
    return git_utils.git_diff('wiki/**/en.md', base_commit=base_commit)


def outdate_translations(*translations, outdated_hash=""):
    """
    Write outdated hash and marker to several translations at once.
    """

    for article_file in translations:
        with open(article_file, "r", encoding='utf-8') as fd:
            front_matter = article_parser.load_front_matter(fd)
        front_matter[OUTDATED_TRANSLATION_TAG] = True
        front_matter[OUTDATED_HASH_TAG] = outdated_hash
        article_parser.save_front_matter(article_file, front_matter)


def check_commit_hashes(modified_translations):
    """
    Validate commit hashes by checking if git has an associated object.
    """

    good_hashes, bad_hashes = set(), set()
    for article_file in modified_translations:
        with open(article_file, "r", encoding='utf-8') as fd:
            front_matter = article_parser.load_front_matter(fd)
        outdated_hash = front_matter.get(OUTDATED_HASH_TAG)
        if outdated_hash is None or outdated_hash in good_hashes:
            continue

        if outdated_hash in bad_hashes:
            yield article_file
            continue

        try:
            git_utils.git("cat-file", "-e", f"{outdated_hash}^{{commit}}")
            good_hashes.add(outdated_hash)
        except RuntimeError:
            bad_hashes.add(outdated_hash)
            yield article_file


def main(*args):
    args = parse_args(args)
    exit_code = 0

    if args.root:
        changed_cwd = file_utils.ChangeDirectory(args.root)

    base_commit = args.base_commit
    if not args.workflow and not base_commit:
        base_commit = git_utils.get_first_branch_commit()

    if not base_commit and not args.all:
        print(f"{console.red('Error:')} neither --base-commit (unable to obtain automatically) nor --all were specified; nothing to do.")
        exit_code = 1
        if args.root:
            del changed_cwd
        return exit_code

    modified_translations = set()
    with_bad_hashes = list()

    if not args.all and base_commit:
        modified_translations = set(list_modified_translations(base_commit))
        with_bad_hashes = list(check_commit_hashes(modified_translations))
    elif args.all:
        all_translations = file_utils.list_all_translations(file_utils.list_all_article_dirs())
        with_bad_hashes = list(check_commit_hashes(all_translations))

    if with_bad_hashes:
        print_bad_hash_error(*with_bad_hashes, outdated_hash=args.outdated_since or base_commit)
        print()
        exit_code = 1

    if not base_commit:
        if args.root:
            del changed_cwd
        return exit_code

    modified_originals = list_modified_originals(base_commit)
    if modified_originals:
        all_translations = file_utils.list_all_translations(sorted(os.path.dirname(tl) for tl in modified_originals))
        translations_to_outdate = list(list_outdated_translations(all_translations, modified_translations))
        if translations_to_outdate:
            outdated_hash = args.outdated_since or base_commit

            should_autofix = getattr(args, AUTOFIX_FLAG[2:], False)
            should_autocommit = getattr(args, AUTOCOMMIT_FLAG[2:], False)
            if should_autofix:
                print(console.green('{} specified, outdating translations...'.format(AUTOFIX_FLAG)))
                outdate_translations(*translations_to_outdate, outdated_hash=outdated_hash)
                if not should_autocommit:
                    print(console.green('Done! To commit the changes, run:'))
                    print(console.green('\tgit add {}; git commit -m "outdate translations"'.format(
                        " ".join(translations_to_outdate)
                    )))
                else:
                    print(console.green('{} specified, committing changes...'.format(AUTOCOMMIT_FLAG)))
                    git_utils.git("add", *translations_to_outdate)
                    git_utils.git("commit", "-m", "outdate translations")
                    print(console.green('Done! The changes have been committed for you.'))
                    print()
                    print(git_utils.git("show", "HEAD", "--no-patch"))
                    print("Changed files:")
                    for file_path in translations_to_outdate:
                        print(console.green(f"* {file_path}"))
            else:
                print_translations_to_outdate(*translations_to_outdate, outdated_hash=outdated_hash, workflow=args.workflow)
                exit_code = 1

        else:
            print(f"{console.grey('Notice:')} all unedited translations are properly outdated.")

    else:
        print(f"{console.grey('Notice:')} no originals are edited, not going to check translations.")

    if args.root:
        del changed_cwd
    return exit_code


if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))



import typing

Redirects = typing.Dict[str, typing.Tuple[str, int]]


def load_redirects(path: str) -> Redirects:
    """
    Read redirects from a YAML dictionary file (done manually to preserve line numbers).
    """

    redirects = {}
    with open(path, 'r', encoding='utf-8') as fd:
        for line_number, line in enumerate(fd, start=1):
            split = line.split('"')
            try:
                redirects[split[1]] = (split[3], line_number)
            except IndexError:
                pass
    return redirects



import typing
from urllib import parse

from wikitools import console


class Reference(typing.NamedTuple):
    """
    The second part of the reference-style link in form of [ref_name]: /wiki/Path/To/Article
    """

    lineno: int
    name: str
    raw_location: str
    parsed_location: parse.ParseResult
    title: str

    def colourise_link(self, fragment_only=False):
        return "{alt_text_in_braces}: {location}{extra}".format(
            alt_text_in_braces=console.green(f"[{self.name}]"),
            location=self.colourise_location(fragment_only=fragment_only),
            extra=f' "{console.blue(self.title)}"' if self.title else "",
        )

    def colourise_location(self, fragment_only=False):
        if fragment_only:
            return "".join((
                console.green(self.parsed_location.path),
                console.red('#' + self.parsed_location.fragment)
            ))
        return console.red(self.raw_location)

    @property
    def start(self):
        return 0

    @property
    def end(self):
        return len(self.name) + 4 + len(self.raw_location) + (
            3 + len(self.title) if self.title else 0
        )


References = typing.Dict[str, Reference]


def extract(s: str, lineno) -> typing.Optional[Reference]:
    """
    Given a line, attempt to extract a reference from it (assuming it occupies the whole line). Example:
        - "[reference]: /wiki/kudosu.png" -> ("reference", "/wiki/kudosu.png")
    """

    if not s.startswith('['):
        return None

    split = s.find(':')
    if not (split != -1 and s[split - 1] == ']' and s[split + 1] == ' '):
        return None

    name = s[1:split - 1]
    try:
        location, title = s[split + 2:].split(' ', maxsplit=1)
        title = title[1:-1]  # trim quotes
    except ValueError:  # no space -> no title
        location = s[split + 2:]
        title = ""

    parsed_location = parse.urlparse(location)
    return Reference(
        lineno=lineno, name=name,
        raw_location=location, parsed_location=parsed_location, title=title
    )


def extract_all(text: str) -> References:
    """
    Attempt to read link references in form of "[reference_name]: /path/to/location" from a text file.
    """

    references = (
        extract(line, lineno)
        for lineno, line in enumerate(text.splitlines(), start=1)
    )
    return {
        r.name: r
        for r in references if
        r is not None
    }



import typing

from wikitools import link_parser

ID_PREFIXES = {'#', 'id='}

ESCAPEABLE_CHARS = { '\\', '(' , ')', '[', ']', '{', '}', '<', '>', '#', '*', '~', '`', '_', '-', '+', '|', '\'', '"', '@', '$', ';', ':', ',', '.' }

# backslashes are sometimes used (perhaps unnecessarily) to avoid remark errors
def unescape(s: str) -> str:
    unescaped = ""
    i = 0
    while i < len(s):
        if s[i] == '\\' and i < len(s) - 1 and s[i + 1] in ESCAPEABLE_CHARS:
            unescaped += s[i + 1]
            i += 2
        else:
            unescaped += s[i]
            i += 1

    return unescaped


def extract_identifier(
    s: str, links_on_line: typing.Optional[typing.List[link_parser.Link]] = None
) -> typing.Tuple[typing.Optional[str], int]:
    """
    Attempt to extract an identifier from a line. These are used in link fragments to skip to certain lines of an article. The following are recognised:
        - Implicit identifiers, which are generated by headings
        - Explicit identifier tags: {#i-den-ti-fi-er} or {id=i-den-ti-fi-er}

    This function doesn't recognise HTML comments and code blocks and will extract identifiers from these as well.
    The burden of checking for the comments and code blocks lies on the caller.
    """

    for i in range(len(s)):
        if s[i] == '{':
            # could be an identifier -- check if there are any meaningful prefixes ahead
            id_start = None
            for prefix in ID_PREFIXES:
                id_start = i + 1 + len(prefix)
                if id_start >= len(s) or s[i + 1: id_start] != prefix:
                    continue

                j = id_start
                while j < len(s) and s[j] != '}':
                    j += 1
                if j < len(s) and s[j] == '}':
                    return (s[id_start: j], id_start)

    # skip regular lines and article titles (no one refers to them)
    if not s.startswith('#') or s.startswith('# '):
        return (None, 0)

    # skip to the actual heading
    j = 0
    while j < len(s) and s[j] in ('#', ' '):
        j += 1

    # headings can contain figures or formatting, but ASC disallows the latter.
    links_on_line = links_on_line or link_parser.find_links(s)

    if not links_on_line:
        heading = s[j:]
    else:
        heading = ""
        start = j
        for k, link in enumerate(links_on_line):
            # before link.start, there is an exclamation mark
            if s[link.start - 1] == '!':
                heading += s[start: link.start - 1]
            else:
                # People/The_Team/Account_support_team has e-mail addresses WITH LINKS IN THEM as headings
                heading += s[start: link.start] + link.alt_text
            start = link.end + 1
            if k == len(links_on_line) - 1:
                heading += s[start:]

    return ("-".join(word.lower() for word in unescape(heading).strip().split()), 0)



import fnmatch
import os
import typing


class ChangeDirectory:
    cwd: str

    def __init__(self, repo_root: str):
        self.cwd = os.getcwd()
        os.chdir(repo_root)

    def __del__(self):
        os.chdir(self.cwd)


def is_newspost(path: str) -> bool:
    return (
        os.path.dirname(path).endswith("news") and
        path.endswith(".md")
    )


def is_article(path: str) -> bool:
    filename = os.path.basename(path)
    return (
        fnmatch.fnmatch(filename, "??.md") or
        fnmatch.fnmatch(filename, "??-??.md")
    )


def is_translation(path: str) -> bool:
    filename = os.path.basename(path)
    return (
        filename != "en.md" and
        (
            fnmatch.fnmatch(filename, "??.md") or
            fnmatch.fnmatch(filename, "??-??.md")
        )
    )


def is_original(path: str) -> bool:
    return os.path.basename(path) == "en.md"


def list_all_files(roots: typing.Iterable[str]=["wiki"]) -> typing.Generator[str, None, None]:
    for item in roots:
        for root, _, filenames in os.walk(item):
            for f in filenames:
                filepath = os.path.join(root, f).replace("\\", "/")
                yield filepath


def list_all_article_dirs() -> typing.Generator[str, None, None]:
    """
    List ALL article directories in the wiki
    """

    for root, _, filenames in os.walk("wiki"):
        if any(is_article(f) for f in filenames):
            yield root.replace("\\", "/")


def list_all_articles() -> typing.Generator[str, None, None]:
    """
    List ALL article files in the wiki
    """

    for filepath in list_all_files(["wiki"]):
        if is_article(filepath):
            yield filepath.replace("\\", "/")


def list_all_newsposts() -> typing.Generator[str, None, None]:
    """
    List ALL newsposts
    """

    for filepath in list_all_files(["news"]):
        if is_newspost(filepath):
            yield filepath.replace("\\", "/")


def list_all_articles_and_newsposts() -> typing.Generator[str, None, None]:
    """
    List ALL article and newspost files 
    """

    for filepath in list_all_files(["wiki", "news"]):
        if is_article(filepath) or is_newspost(filepath):
            yield filepath.replace("\\", "/")


def list_all_translations(article_dirs: typing.Iterable[str]) -> typing.Generator[str, None, None]:
    """
    List ALL translations inside the articles specified
    """

    for d in article_dirs:
        for filename in sorted(os.listdir(d)):
            if is_original(filename) or not is_article(filename):
                continue

            yield os.path.join(d, filename).replace("\\", "/")



import collections
import typing

from wikitools import console, link_parser, reference_parser


class LinkError:
    _colourise_fragment_only: bool = False
    link: link_parser.Link

    def pretty(self):
        return f'{console.blue("Note:")} {repr(self)}'

    def pretty_location(self, article_path, lineno):
        return "{}: {}".format(
            console.yellow(":".join((article_path, str(lineno), str(self.pos)))),
            self.link.colourise_location(fragment_only=self._colourise_fragment_only)
        )

    @property
    def pretty_link(self):
        return self.link.colourise_link(fragment_only=self._colourise_fragment_only)

    @property
    def pos(self):
        return self.link.start + 1


class MalformedLinkError(
    LinkError,
    collections.namedtuple('MalformedLinkError', 'link')
):
    """
    An error indicating an erroneous link (for example, with several leading slashes).
    """

    link: link_parser.Link

    def __repr__(self):
        return f'Incorrect link structure (typo?): "{self.link.raw_location}"'


class LinkNotFoundError(
    LinkError,
    collections.namedtuple('LinkNotFound', 'link reference resolved_location')
):
    """
    An error indicating a missing link: a text or binary file does not exist, and there is no redirect for it.
    """

    link: link_parser.Link
    reference: typing.Optional[reference_parser.Reference]
    resolved_location: str

    def __repr__(self):
        return '"{}" was not found {}'.format(
            self.resolved_location,
            f"(reference at line {self.reference.lineno})"
            if self.reference else ''
        )


class BrokenRedirectError(
    LinkError,
    collections.namedtuple('BrokenRedirect', 'link resolved_location redirect_lineno redirect_destination')
):
    """
    An error indicating broken redirect: the redirect either points to a non-existent article, or another redirect (which is not allowed).
    """

    link: link_parser.Link
    resolved_location: str
    redirect_lineno: int
    redirect_destination: str

    def __repr__(self):
        return 'Broken redirect (redirect.yaml:{}: {} --> {})'.format(
            self.redirect_lineno, self.resolved_location.lower(), self.redirect_destination
        )


class MissingReferenceError(
    LinkError,
    collections.namedtuple('MissingReference', 'link')
):
    """
    An error indicating that a reference-style link is missing its counterpart:
    [link][link_ref] exists, but [link_ref]: /wiki/Path/To/Article does not.
    """

    link: link_parser.Link

    def __repr__(self):
        return f'No corresponding reference found for "{self.link.raw_location}"'


class MissingIdentifierError(
    LinkError,
    collections.namedtuple('MissingIdentifier', 'link path identifier no_translation_available translation_outdated')
):
    """
    An error indicating that in another article there is no heading or identifier tag
    that would produce #such-reference.
    """

    _colourise_fragment_only = True

    link: link_parser.Link
    path: str
    identifier: str
    # for news posts, these two should be False
    no_translation_available: bool # also implies a link to and from a translation
    translation_outdated: bool

    def __repr__(self):
        return 'There is no heading or tag with identifier "{}" in "/{}"{}'.format(
            self.identifier, self.path,
            ' (no translation available)' if self.no_translation_available
            else ' (outdated translation)' if self.translation_outdated
            else ''
        )

    @property
    def pos(self):
        return self.link.fragment_start + 1



import collections
import io
import pathlib
import shutil
import typing

import yaml

from wikitools import code_block_parser, link_parser, comment_parser, identifier_parser, reference_parser

FRONT_MATTER_DELIMITER = '---'
TITLE_INDICATOR = '# '


# Workaround to make yaml.Dumper write lists with leading indentation
# (taken from https://github.com/yaml/pyyaml/issues/234#issuecomment-765894586)
class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


class ArticleLine(typing.NamedTuple):
    raw_line: str
    links: typing.List[link_parser.Link]


class Article:
    """
    A wiki article, which contains some parts from the text file considered important:
        - Only lines containing links, with the full list of parsed links
        - List of references for reference-style links, usually found at the bottom of the article
        - List of identifiers, which #can-be-referred-to from other articles
    """

    directory: str
    filename: str
    lines: typing.Dict[int, ArticleLine]
    references: reference_parser.References
    identifiers: typing.Dict[str, int]
    front_matter: dict

    def __init__(
        self, path: pathlib.Path,
        lines: typing.Dict[int, ArticleLine],
        references: reference_parser.References,
        identifiers: typing.Dict[str, int],
        front_matter: dict
    ):
        self.filename = path.name
        self.directory = str(path.parent.as_posix())
        self.lines = lines
        self.references = references
        self.identifiers = identifiers
        self.front_matter = front_matter

    @property
    def path(self) -> str:
        return '/'.join((self.directory, self.filename))


def load_front_matter(fileobj: typing.TextIO) -> dict:
    delimiters = 0
    buffer = io.StringIO()

    offset = fileobj.tell()
    for line in fileobj:
        if line.split('#')[0].strip() == FRONT_MATTER_DELIMITER:
            delimiters += 1
        # Stop on second delimiter, or when it's clear there won't be front matter at all
        if delimiters == 2 or line.startswith(TITLE_INDICATOR):
            break
        buffer.write(line)
    fileobj.seek(offset)

    if delimiters == 2:
        return yaml.safe_load(buffer.getvalue())
    return dict()


class FrontMatterDetector():
    """
    Helper class for keeping track of whether a line is part of front matter.
    Includes delimiter lines.
    """

    def __init__(self):
        self.in_front_matter_ = 0

    in_front_matter_ = False

    def in_front_matter(self, line: str):
        if line.split('#')[0].strip() == FRONT_MATTER_DELIMITER:
            self.in_front_matter_ = not self.in_front_matter_
            return True
        return self.in_front_matter_


def save_front_matter(filepath: str, fm: dict):
    if not fm:
        return

    new_path = filepath + '.new'
    with open(new_path, 'w', encoding='utf-8') as new_file:
        new_file.write(FRONT_MATTER_DELIMITER + '\n')
        new_file.write(yaml.dump(
            fm, Dumper=Dumper, default_flow_style=False, indent=2, sort_keys=False, allow_unicode=True,
        ))
        new_file.write(FRONT_MATTER_DELIMITER + '\n\n')

        with open(filepath, "r", encoding='utf-8') as old_file:
            front_matter_detector = FrontMatterDetector()
            for line in old_file:
                # There shouldn't be anything before any front matter, so it's just assumed here
                # There may be cases where the title is preceded by a comment or HTML tags, however
                # Such content is preserved
                if not front_matter_detector.in_front_matter(line) and line.strip():
                    new_file.write(line)
                    new_file.write(old_file.read())
                    break

    shutil.move(new_path, filepath)


def parse(path: typing.Union[str, pathlib.Path]) -> Article:
    """
    Read an article line by line, extracting links, identifiers and references as we go.
    Anything inside <!-- HTML comments -->, both single and multiline, is skipped.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    saved_lines = {}
    references = {}
    cnt: typing.Counter[str] = collections.Counter()
    identifiers: typing.Dict[str, int] = {}

    comment_reader = comment_parser.CommentParser()
    code_block_reader = code_block_parser.CodeBlockParser()
    with path.open('r', encoding='utf-8') as fd:
        front_matter = load_front_matter(fd)
        for lineno, line in enumerate(fd, start=1):
            comments = comment_reader.parse(line)
            code_blocks = code_block_reader.parse(line)

            # everything in a multiline comment or code block doesn't count
            if comment_reader.in_multiline or code_block_reader.in_multiline:
                continue

            links_on_line = list(filter(
                lambda l: not(
                    l.content == '/wiki/Sitemap' or
                    comment_parser.is_in_comment(l.start, comments) or
                    code_block_parser.is_in_code_block(l.start, code_blocks)
                ),
                link_parser.find_links(line)
            ))
            # cache meaningful lines. in our case, lines with links and references
            if links_on_line:
                saved_lines[lineno] = ArticleLine(raw_line=line, links=links_on_line)

            identifier, pos = identifier_parser.extract_identifier(line, links_on_line)
            # if a comment contains identifiers, this assumes such a comment at least
            # doesn't appear before an actual identifier. this is a rare occurrence anyway
            if identifier is not None and not comment_parser.is_in_comment(pos, comments):
                    cnt[identifier] += 1
                    # duplicate identifiers get a suffix
                    if identifier in identifiers:
                        identifier = '{}.{}'.format(identifier, cnt[identifier] - 1)
                    identifiers[identifier] = lineno

            if line.startswith('['):
                reference = reference_parser.extract(line.strip(), lineno=lineno)
                if reference is not None:
                    references[reference.name] = reference

    return Article(path, saved_lines, references, identifiers, front_matter)



#!/usr/bin/env python3

import argparse
import sys
import typing

from wikitools import article_parser, console, link_checker, redirect_parser, errors as error_types, file_utils


def print_error():
    print(f"{console.red('Error:')} Some wiki or image links in the files you've changed have errors.\n")
    print("This can happen in one of the following ways:\n")
    print("- The article or image that the link points to has since been moved or renamed (make sure to match capitalisation)")
    print("- The link simply contains typos or formatting errors")
    print("- The link works, but contains locale selection (e.g. /wiki/en/Article_styling_criteria instead of /wiki/Article_styling_criteria)")
    print("- The link works, but contains URL-escaped characters (https://en.wikipedia.org/wiki/Percent-encoding). This only applies for links to articles and images inside the wiki.")
    print("- The link works, but incurs multiple redirects. Use a direct link instead.")
    print("\nFor more information on link style, see https://osu.ppy.sh/wiki/en/Article_styling_criteria/Formatting#links.")
    print(f"\nIf you need to bypass this check, add {console.red('SKIP_WIKILINK_CHECK')} anywhere in the PR description.\n")


def print_clean():
    print("Notice: No broken wiki or image links detected.")


def s(i: int, s: str) -> str:
    return f"{i} {s}{'s' if i != 1 else ''}"


def print_count(errors: int, matches: int, error_files: int, files: int):
    print(f"{console.blue('Note:')} Found {s(errors, 'error')} in {s(error_files, 'file')} ({s(matches, 'link')} in {s(files, 'file')} checked).")


def highlight_links(s: str, errors: typing.List[error_types.LinkError]) -> str:
    highlighted_line = ""
    prev_index = 0
    for error in errors:
        highlighted_line += s[prev_index: error.link.start]
        highlighted_line += error.pretty_link
        prev_index = error.link.end + 1
    highlighted_line += s[prev_index: -1]
    return highlighted_line


def pretty_location(path, lineno, pos, location):
    return f"{console.yellow(path)}:{lineno}:{pos}: {console.red(location)}"


def parse_args(args):
    parser = argparse.ArgumentParser(usage="%(prog)s check-links [options]")
    parser.add_argument("-t", "--target", nargs='*', help="paths to the articles you want to check, relative to the repository root")
    parser.add_argument("-a", "--all", action='store_true', help="check all articles")
    parser.add_argument("-s", "--separate", action='store_true', help="print errors that appear on the same line separately")
    parser.add_argument("--outdated", action='store_true', help="check links in outdated articles and section links to outdated translations")
    parser.add_argument("-r", "--root", help="specify repository root, current working directory assumed otherwise")
    return parser.parse_args(args)


def identifier_suggestions(e, articles):
    return '\n\t'.join((
        'line {}: {}'.format(lineno, identifier)
        for identifier, lineno in sorted(
            articles[e.path].identifiers.items(), key=lambda tuple_: tuple_[1]
        )
    ))


def main(*args):
    args = parse_args(args)
    if not args.target and not args.all:
        print(f"{console.grey('Notice:')} No articles to check.")
        sys.exit(0)

    if args.root:
        changed_cwd = file_utils.ChangeDirectory(args.root)

    filenames = []
    if args.all:
        filenames = file_utils.list_all_articles_and_newsposts()
    else:
        filenames = list(filter(lambda x: file_utils.is_article(x) or file_utils.is_newspost(x), args.target))

    redirects = redirect_parser.load_redirects("wiki/redirect.yaml")
    exit_code = 0

    articles = {}
    for filename in filenames:
        a = article_parser.parse(filename)
        articles[a.path] = a

    error_count = 0
    link_count = 0
    error_file_count = 0
    file_count = 0

    for _, a in sorted(articles.items()):
        if not args.outdated and (a.front_matter.get("outdated", False) or a.front_matter.get("outdated_translation", False)):
            continue

        link_count += sum(len(_.links) for _ in a.lines.values())
        file_count += 1

        errors = link_checker.check_article(a, redirects, articles)

        if not args.outdated:
            errors = {
                a: b for a, b in {
                    i: [
                        e for e in errors_on_line if
                        not (
                            isinstance(e, error_types.MissingIdentifierError)
                            and (e.no_translation_available or e.translation_outdated)
                        )
                    ]
                    for i, errors_on_line in errors.items()
                }.items()
                if b
            }

        if not errors:
            continue

        error_file_count += 1
        if exit_code == 0:
            print_error()
        exit_code = 1

        for lineno, errors_on_line in sorted(errors.items()):
            error_count += len(errors_on_line)
            for e in errors_on_line:
                print(e.pretty_location(a.path, lineno))
            for e in errors_on_line:
                print(e.pretty())
                if isinstance(e, error_types.MissingIdentifierError):
                    suggestions = identifier_suggestions(e, articles)
                    if suggestions:
                        print('{}\n\t{}'.format(console.blue('Suggestions:'), suggestions))

            print()
            if args.separate:
                for e in errors_on_line:
                    print(highlight_links(a.lines[lineno].raw_line, [e]), end="\n\n")
            else:
                print(highlight_links(a.lines[lineno].raw_line, errors_on_line), end="\n\n")

    if exit_code == 0:
        print_clean()
        print()

    print_count(error_count, link_count, error_file_count, file_count)
    if args.root:
        del changed_cwd
    return exit_code


if __name__ == "__main__":
    sys.exit(main(*sys.argv[1:]))



import subprocess as sp


def git(*args, expected_code=0):
    cmd = ["git"] + list(map(str, args))
    proc = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = proc.communicate()
    out = out.decode("utf-8") if out else ""
    err = err.decode("utf-8") if err else ""
    if proc.returncode != expected_code:
        raise RuntimeError(
            "{} failed:\n"
            "- exit code: {}\n"
            "- stdout: {!r}\n"
            "- stderr: {!r}\n".format(
                cmd, proc.returncode, out, err
            )
        )
    return out


def git_diff(*file_paths, base_commit=""):
    res = git("diff", "--diff-filter=d", "--name-only", f"{base_commit}^", "--", *file_paths)
    return res.splitlines()


def get_first_branch_commit():
    """
    Gets the first commit of the current branch where it diverges from master.

    Returns None if already on master, or if the feature branch is already merged into master

    If the current branch is more up to date with upstream master than local master, an even earlier commit will be returned. This can happen when checking out a PR branch; update local master with upstream if so.
    """

    res = git("log", "master..", "--pretty=format:%H").splitlines()
    return res[-1] if res else None




import os
import sys
import typing

article_path = "wiki/Tournaments/Countries_that_participated_in_osu!_tournaments"


class Section:
    IDLE = 0
    GENERAL = 1
    OSU = 2
    TAIKO = 3
    CATCH = 4
    MANIA = 5
    WC_OSU = 6
    WC_TAIKO = 7
    WC_CATCH = 8
    WC_MANIA = 9

sections = {
    "### General ranking": Section.GENERAL,
    "### ![][osu!] osu! ranking": Section.OSU,
    "### ![][osu!taiko] osu!taiko ranking": Section.TAIKO,
    "### ![][osu!catch] osu!catch ranking": Section.CATCH,
    "### ![][osu!mania] osu!mania ranking": Section.MANIA,
    "### ![][osu!] osu! World Cup": Section.WC_OSU,
    "### ![][osu!taiko] osu!taiko World Cup": Section.WC_TAIKO,
    "### ![][osu!catch] osu!catch World Cup": Section.WC_CATCH,
    "### ![][osu!mania] osu!mania World Cup": Section.WC_MANIA,
}

tables: typing.Dict[int, typing.List[str]] = {
    Section.GENERAL: [],
    Section.OSU: [],
    Section.TAIKO: [],
    Section.CATCH: [],
    Section.MANIA: [],
    Section.WC_OSU: [],
    Section.WC_TAIKO: [],
    Section.WC_CATCH: [],
    Section.WC_MANIA: [],
}

with open(article_path + "/en.md", 'r', encoding='utf-8') as fd:
    section = Section.IDLE
    for line in fd:
        if line.startswith('#') and line.strip() in sections:
            section = sections[line.strip()]
            continue
        elif section == Section.IDLE:
            continue

        print(line)
        if line.startswith('|'):
            tables[section].append(line)

for table in tables.values():
    print(''.join(table))

#for f in os.listdir(article_path):
#    if not f.endswith(".md") or f.endswith("en.md"):
#        continue
#    
#    for line in 



# type: ignore
import setuptools

from wikitools_cli.VERSION import VERSION

with open("README.md", "r", encoding="utf-8") as fd:
    long_description = fd.read()

setuptools.setup(
    name="osu-wiki-tools",
    version=VERSION,
    author="Walavouchey",
    author_email="wala@yui.tv",
    description="Various tools for osu! wiki contributors",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Walavouchey/osu-wiki-tools",
    project_urls={
        "Bug Tracker": "https://github.com/Walavouchey/osu-wiki-tools/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    test_suite="tests",
    packages=setuptools.find_packages(exclude=["tests", "tests.visual"]),
    entry_points={
        "console_scripts": [
            "osu-wiki-tools=wikitools_cli.osu_wiki_tools:console_main",
        ],
    },
    python_requires=">=3.10",
    install_requires=[
        "PyYAML==6.0",
        "types-PyYAML==6.0.6",
        "yamllint==1.26.3",
    ],
)



#!/usr/bin/env python3

"""
Tools useful for osu! wiki contributors
"""

import argparse
import sys

from wikitools_cli.commands import check_outdated_articles, check_links, check_yaml

from wikitools_cli.VERSION import VERSION

commands = [
    {
        "name": "check-outdated-articles",
        "help": "check if articles are correctly outdated",
        "entry": check_outdated_articles.main,
    },
    {
        "name": "check-links",
        "help": "find broken wikilinks",
        "entry": check_links.main,
    },
    {
        "name": "check-yaml",
        "help": "validate front matter and standalone YAML files",
        "entry": check_yaml.main,
    },
]


def split_args(args):
    main_args = []
    subcommand_args = []
    found_command_name = False

    for arg in args:
        if found_command_name:
            subcommand_args.append(arg)
        else:
            main_args.append(arg)
        if any(arg == command["name"] for command in commands):
            found_command_name = True

    return main_args, subcommand_args


def parse_args(args):
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-V", "--version", action="version", version=VERSION)

    subparsers = parser.add_subparsers(title="commands", metavar="command")
    for command in commands:
        subparser = subparsers.add_parser(command["name"], help=command["help"], add_help=False)
        subparser.set_defaults(func=command["entry"])

    # any parameters after the command name are validated in the command's respective file
    # this main parser would error on them as unrecognised, so they're cut away here
    main_args, subcommand_args = split_args(args)

    parsed_args = parser.parse_args(main_args)
    if len(vars(parsed_args)) == 0:
        # program was invoked without any arguments
        parser.print_help()
        sys.exit(1)

    return parsed_args, subcommand_args


def main(*args):
    parsed_args, subcommand_args = parse_args(args)
    return parsed_args.func(*subcommand_args)


def console_main():
    sys.exit(main(*sys.argv[1:]))


if __name__ == '__main__':
    console_main()



import os
import pathlib
import typing

from wikitools import redirect_parser, reference_parser, errors, link_parser, article_parser
from wikitools import console


def check_link(
    article: article_parser.Article, link: link_parser.Link,
    redirects: redirect_parser.Redirects, references: reference_parser.References,
    all_articles: typing.Dict[str, article_parser.Article]
) -> typing.Optional[errors.LinkError]:
    """
    Verify that the link is valid:
        - External links are always assumed valid, since we can't just issue HTTP requests left and right
        - For Markdown references, there exists a dereferencing line with [reference_name]: /lo/ca/ti/on
        - Direct internal links, as well as redirects, must point to existing article files
        - Relative links are parsed under the assumption that they are located inside the current article's directory
    """

    # resolve the link, if possible
    reference = link.resolve(references)
    if reference is None and link.is_reference:
        return errors.MissingReferenceError(link)

    location = reference.parsed_location.path if reference else link.parsed_location.path
    parsed_location = reference.parsed_location if reference else link.parsed_location

    # news post link
    if ((parsed_location.scheme == "http" or parsed_location.scheme == "https") and
        parsed_location.netloc == "osu.ppy.sh" and location.startswith("/home/news/")):
        target = pathlib.Path(location[1:] + ".md").relative_to("home")
        year = target.name.split("-")[0]
        repo_target = pathlib.Path(f"news/{year}/{target.name}")
        location = '/' + repo_target.as_posix()

        if not repo_target.exists():
            # news posts don't have redirects
            return errors.LinkNotFoundError(link, reference, location)
        else:
            if not parsed_location.fragment:
                return None

            raw_path = repo_target.as_posix()
            if raw_path not in all_articles:
                all_articles[raw_path] = article_parser.parse(raw_path)
            target_article = all_articles[raw_path]

            if parsed_location.fragment not in target_article.identifiers:
                return errors.MissingIdentifierError(link, raw_path, parsed_location.fragment, False, False)
            else:
                return None

    # some external link; don't care
    if parsed_location.scheme:
        return None

    # domain is non-empty, but the link is internal?
    if parsed_location.netloc:
        return errors.MalformedLinkError(link)

    # convert a relative wikilink to absolute
    if not location.startswith("/wiki/"):
        current_article_dir = os.path.relpath(article.directory, 'wiki')
        location = f"/wiki/{current_article_dir}/{location}"

    target = pathlib.Path(location[1:])  # strip leading slash
    # no article? could be a redirect
    if not target.exists():
        redirect_source = target.relative_to('wiki').as_posix()
        try:
            redirect_destination, redirect_line_no = redirects[redirect_source.lower()]
        except KeyError:
            return errors.LinkNotFoundError(link, reference, location)

        target = pathlib.Path('wiki') / redirect_destination
        if not target.exists():
            return errors.BrokenRedirectError(link, redirect_source, redirect_line_no, redirect_destination)

    # link to an article in general, article exists -> good
    if not parsed_location.fragment:
        return None

    # link to a section -> need to find the target article; it could be a translation
    # XXX(TicClick): this part assumes there is always an English version of the article in a folder
    target_file = target / article.filename
    translation = target_file # verified to be the case later
    no_translation_available = not (article.filename != 'en.md' and target_file.is_file())

    if no_translation_available:
        target_file = target / 'en.md'

    raw_path = target_file.as_posix()
    if raw_path not in all_articles:
        # this is safe to do since the caller iterates over a copy of all_articles -> we can modify it as we wish
        all_articles[raw_path] = article_parser.parse(target_file)
    target_article = all_articles[raw_path]

    if parsed_location.fragment not in target_article.identifiers:
        # collect some additional metadata before reporting
        translation_outdated = False
        if not no_translation_available:
            raw_path_translation = translation.as_posix()
            if raw_path_translation not in all_articles:
                # this is safe to do since the caller iterates over a copy of all_articles -> we can modify it as we wish
                all_articles[raw_path_translation] = article_parser.parse(translation)
            translation_outdated = all_articles[raw_path_translation].front_matter.get('outdated_translation', False)

        return errors.MissingIdentifierError(
            link, raw_path, parsed_location.fragment,
            no_translation_available, translation_outdated
        )

    return None


def check_article(
    article: article_parser.Article, redirects: redirect_parser.Redirects,
    all_articles: typing.Dict[str, article_parser.Article]
) -> typing.Dict[int, typing.List[errors.LinkError]]:
    """
    Try resolving links in the article to other articles or files.
    """

    errors = {}
    for lineno, line in article.lines.items():
        local_errors = [
            errors for errors in (
                check_link(article, link, redirects, article.references, all_articles)
                for link in line.links
            )
            if errors
        ]
        if local_errors:
            errors[lineno] = local_errors

    return errors



import abc
import typing

import yaml
import yamllint.rules  # type: ignore


ALLOWED_FRONT_MATTER_TAGS = frozenset({
    # Article tags
    "needs_cleanup",
    "layout",  # Main Page uses this
    "legal",
    "outdated",
    "outdated_since",
    "outdated_translation",
    "stub",
    "tags",
    "translate_from",  # https://github.com/ppy/osu-wiki/pull/7865/commits/ba6a169add4f3cac620e0147e81323a28cd27376
    "no_native_review",

    # Newspost tags
    "date",
    "layout",
    "title",
    "tumblr_url",
})


class _JunkMatcher:
    def match_file(self, p: str):
        return not p.endswith(".md")


class _State(list):
    def start_sequence(self):
        self.append(yaml.SequenceStartEvent(None, None, False))

    def start_mapping(self):
        self.append(yaml.MappingStartEvent(None, None, False))

    def end_nested_object(self):
        if len(self) > 0:
            self.pop()

    def inside_sequence(self):
        return len(self) > 0 and isinstance(self[-1], yaml.SequenceStartEvent)

    def inside_mapping(self):
        return len(self) > 0 and isinstance(self[-1], yaml.MappingStartEvent)


class _FrontMatterRule(dict, metaclass=abc.ABCMeta):
    TYPE = "token"
    LEVEL = "error"

    @property
    @abc.abstractmethod
    def ID(self):
        raise ValueError("Missing rule identifier")

    @abc.abstractmethod
    def inner_check(
        self, state: _State, prev_token: yaml.Token, token: yaml.Token,
        next_token: yaml.Token, next_next_token: yaml.Token
    ):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['ignore'] = _JunkMatcher()
        self['level'] = self.LEVEL

    @staticmethod
    def _is_start_of_mapping(token: yaml.Token):
        return isinstance(token, (yaml.BlockMappingStartToken, yaml.FlowMappingStartToken))

    @staticmethod
    def _is_start_of_sequence(token: yaml.Token):
        return isinstance(token, (yaml.BlockSequenceStartToken, yaml.FlowSequenceStartToken))

    @staticmethod
    def _is_end_of_nested_block(token: yaml.Token):
        return isinstance(token, (yaml.BlockEndToken, yaml.FlowMappingEndToken, yaml.FlowSequenceEndToken))

    @staticmethod
    def _is_mapping_key(prev_token: yaml.Token, token: yaml.Token, next_token: yaml.Token):
        return (
            isinstance(prev_token, yaml.KeyToken) and isinstance(token, yaml.ScalarToken) and
            isinstance(next_token, yaml.ValueToken)
        )

    @classmethod
    def _make_problem(cls, next_token: yaml.Token, issue: str):
        return yamllint.linter.LintProblem(
            next_token.start_mark.line + 1, next_token.start_mark.column + 1, issue, cls.LEVEL
        )

    def check(
        self, conf: yamllint.config.YamlLintConfig, token: yaml.Token,
        prev_token: yaml.Token, next_token: yaml.Token, next_next_token: yaml.Token, context: dict
    ):
        state = context.setdefault("state", _State())
        error = self.inner_check(state, prev_token, token, next_token, next_next_token)
        if error is not None:
            yield error

        if self._is_start_of_mapping(token):
            state.start_mapping()
        elif self._is_start_of_sequence(token):
            state.start_sequence()
        elif self._is_end_of_nested_block(token):
            state.end_nested_object()


class NestedStructureRule(_FrontMatterRule):
    ID = "osu-wiki-nested-structure"

    def inner_check(
        self, state: _State, prev_token: yaml.Token, token: yaml.Token,
        next_token: yaml.Token, next_next_token: yaml.Token
    ):
        # "tags", the top-level list of article tags, is the only field allowed to contain lists
        allowed_combination = state.inside_mapping() and self._is_start_of_sequence(token)
        if (
            (state.inside_mapping() or state.inside_sequence()) and
            (self._is_start_of_mapping(token) or self._is_start_of_sequence(token)) and
            not allowed_combination
        ):
            return self._make_problem(
                next_token,
                "bad front matter: lists or dictionaries cannot contain other similarly complex objects"
            )


class TopLevelRule(_FrontMatterRule):
    ID = "osu-wiki-top-level"

    def inner_check(
        self, state: _State, prev_token: yaml.Token, token: yaml.Token,
        next_token: yaml.Token, next_next_token: yaml.Token
    ):
        if self._is_start_of_sequence(token) and len(state) == 0:
            return self._make_problem(next_token, "bad front matter: the top level must be a dictionary, not a list")


class AllowedTagsRule(_FrontMatterRule):
    ID = "osu-wiki-allowed-tags"

    def inner_check(
        self, state: _State, prev_token: yaml.Token, token: yaml.Token,
        next_token: yaml.Token, next_next_token: yaml.Token
    ):
        if self._is_mapping_key(prev_token, token, next_token):
            value = token.value  # type: ignore
            if value not in ALLOWED_FRONT_MATTER_TAGS:
                return self._make_problem(token, f"bad front matter: {value!r} is not in the list of allowed tags")


OSU_WIKI_RULES: typing.List[typing.Type[_FrontMatterRule]] = [
    NestedStructureRule,
    TopLevelRule,
    AllowedTagsRule,
]
